/****************************************************************
 *    	Analysis of Quasi-Elastic e-p & e-n Q^2   		  		*
 *		Version 6 | 18-jul-12									*
 *   	Alexander Balsamo: alexander.balsamo@richmond.edu  		*
 *		Coatjava Version: 	5a.3.3	| Released: 18-jun-01  		*
 * **************************************************************
 ~																~
 * **************************************************************
 *	How to Run: 												*
 *	~$ /run-groovy epenQ2v6.groovy yourReconstructedFile.hipo 	*
 *																*
 * This will output a histogram hipo file: histEpenQ2.hipo 		*
 * 	which can be viewed using broswer in Coatjava  				*
 * 	or epenQ2HistViewer.groovy 									*
 ****************************************************************/


// imports ##############################################################################################################
// for reconstruction data
import org.jlab.io.evio.*;
import org.jlab.io.hipo.*;

// for histograms
import org.jlab.groot.data.TDirectory;
import org.jlab.groot.data.H1F;
import org.jlab.groot.data.H2F;
import org.jlab.groot.ui.TCanvas;

// for vectors
import org.jlab.geom.prim.Vector3D;

// for Geometry and Swimming
import org.jlab.clas.physics.Particle;
import org.jlab.detector.base.GeometryFactory;
import org.jlab.detector.base.DetectorType;
import org.jlab.clas.fastmc.*;
import org.jlab.geom.prim.Line3D;
import org.jlab.geom.prim.Shape3D;
import org.jlab.geom.prim.Plane3D;
import org.jlab.geom.prim.Point3D;
import org.jlab.geom.detector.ec.ECSector;
import org.jlab.geom.detector.ec.ECSuperlayer;
import org.jlab.geom.detector.ec.ECLayer;
import org.jlab.geom.base.ConstantProvider;
import org.jlab.geom.detector.ec.ECFactory;
import org.jlab.geom.detector.ec.ECDetector;

// for Objects
import java.lang.Math;
import java.util.ArrayList;
import java.text.DecimalFormat;

// class declaration ####################################################################################################
class QuasiElastic{
	// variable declarations
	final double BEAM_ENERGY; // GeV
	final double TARGET_MASS; // au (atomic mass units)
	final double NUCLEON_MASS; // au

	DecimalFormat numberFormat = new DecimalFormat("#.00000"); // will only keep up to 5 decimal points when applied

	// Swimming Particles ***********************************************************************************************
	// for protons 
	Particle calcProton;
	Clas12FastMC recFastMC = new Clas12FastMC(-1.0,1.0); //torus scale, solenoid scale

	String[] negRequirmentDC = ["DC"];
	int[] negRequirmentDC_Layers = [36];
	String[] negRequirmentEC = ["EC"];
	int[] negRequirmentEC_Layers = [9];
	String[] negRequirmentFTOF = ["FTOF"];
	int[] negRequirmentFTOF_Layers = [1];

	// for neutrons 
	Line3D swumNeutronPath;
	ArrayList sectors, superlayers, layers;
	ECSector sector;
	ECSuperlayer superlayer;
	ECLayer layer;
	Shape3D[] ECfaces = new Plane3D[6];
	boolean intersectionStatus = false;

	// create histograms ************************************************************************************************
	TDirectory histFile = new TDirectory();
	H1F q2Raw;
	H1F q2Acceptance;
	H1F q2ep;
	H1F q2epThetaPq;
	H1F q2en;
	H1F q2enThetaPq;
	H1F q2epSim;
	H1F q2enSim;

	// constructer for methods ##########################################################################################
	QuasiElastic(){
		// finals
		BEAM_ENERGY = 11.0; // GeV
		TARGET_MASS = 2.014; // au (atomic mass units)
		NUCLEON_MASS = 1.0; // au

		// setup for proton swim, Detector Elements, and FastMC *********************************************************
		recFastMC.addDetector("DC", GeometryFactory.getDetector(DetectorType.DC));
		recFastMC.addDetector("ECAL", GeometryFactory.getDetector(DetectorType.ECAL));
		recFastMC.addDetector("FTOF", GeometryFactory.getDetector(DetectorType.FTOF));

		recFastMC.addFilter(1, negRequirmentDC, negRequirmentDC_Layers); // (charge, detector, minimum number of hits)
		recFastMC.addFilter(1, negRequirmentEC, negRequirmentEC_Layers);
		recFastMC.addFilter(1, negRequirmentFTOF, negRequirmentFTOF_Layers);

		// setup for neutron swim, Detector Elements ********************************************************************
		ConstantProvider constants = GeometryFactory.getConstants(DetectorType.ECAL,11,"default");
		ECDetector detector = new ECFactory().createDetectorCLAS(constants);

		sectors = detector.getAllSectors();

		// loop over sectors moving down the hierarchy to get a plane in the face of each EC Sector
		for (int i = 0; i < 6; i++){
		    sector = sectors.get(i);
		    superlayers = sector.getAllSuperlayers();
		    superlayer = superlayers.get(0);
		    layers = superlayer.getAllLayers();
		    layer = layers.get(0);
		    ECfaces[i] = layer.getBoundary();
		} // end loop over sectors

		// histograms ***************************************************************************************************
		histFile.mkdir("epenQ2");

		q2Raw = new H1F("q2Raw",60,0,15);
		q2Raw.setFillColor(33);
		q2Raw.setTitle("Q^2 Raw");
		q2Raw.setTitleX("Q^2");

		q2Acceptance = new H1F("q2Acceptance",60,0,15);
		q2Acceptance.setFillColor(32);
		q2Acceptance.setTitle("Q^2 Accepted");
		q2Acceptance.setTitleX("Q^2");

		q2ep = new H1F("q2ep",60,0,15);
		q2ep.setFillColor(39);
		q2ep.setTitle("Q^2 e-p Events");
		q2ep.setTitleX("Q^2");

		q2epThetaPq = new H1F("q2epThetaPq",60,0,15);
		q2epThetaPq.setFillColor(39);
		q2epThetaPq.setTitle("Q^2 e-p ThetaPq Cut");
		q2epThetaPq.setTitleX("Q^2");

		q2en = new H1F("q2en",60,0,15);
		q2en.setFillColor(38);
		q2en.setTitle("Q^2 e-n Events");
		q2en.setTitleX("Q^2");

		q2enThetaPq = new H1F("q2enThetaPq",60,0,15);
		q2enThetaPq.setFillColor(38);
		q2enThetaPq.setTitle("Q^2 e-n ThetaPq Cut");
		q2enThetaPq.setTitleX("Q^2");

		q2epSim = new H1F("q2epSim",60,0,15);
		q2epSim.setFillColor(33);
		q2epSim.setTitle("Q^2 e-p Simulated");
		q2epSim.setTitleX("Q^2");

		q2enSim = new H1F("q2enSim",60,0,15);
		q2enSim.setFillColor(33);
		q2enSim.setTitle("Q^2 e-n Simulated");
		q2enSim.setTitleX("Q^2");
	} // end of constructor

	// Main Method ######################################################################################################
	public static void main(String[] args){
		// creates instance of class and import file ********************************************************************
		QuasiElastic epen = new QuasiElastic();

		// First argument is file name
		String inputFile = args[0];

		// Create object to read in the data file
		HipoDataSource reader = new HipoDataSource();
		reader.open(inputFile);
		System.out.println("-----------> Reading File: [" + inputFile + "]\n");

		// create an event
		HipoDataEvent event;

		// counts *******************************************************************************************************
		int nEvents = 1; // start at first event
		int acceptedEvents = 0;
		int foundProton = 0;
		int foundNeutron = 0;

		// loop over all events $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
		System.out.println("-----------> X gon' give it to ya...\n");
		while(reader.hasEvent()){

			// get the event
			event = reader.getNextEvent();

			if(epen.checkList(event) && epen.acceptanceMatching(event.getBank("TimeBasedTrkg::TBTracks")) ){
				acceptedEvents++;

				// ep Block
				if(epen.epThetaPq(event.getBank("TimeBasedTrkg::TBTracks"))){
					foundProton++;

					// look for simulated event to compare with found data
					epen.epenSim(event.getBank("MC::Particle"));
				}

				// en Block
				if(epen.enThetaPq(event.getBank("TimeBasedTrkg::TBTracks"), event.getBank("ECAL::clusters"))){
					foundNeutron++;

					// look for simulated event to compare with found data
					epen.epenSim(event.getBank("MC::Particle"));
				}
			} // end of if checkList and acceptanceMatching

			// Look for next event in the while loop
			nEvents++;
		} // end the while loop

		// add histograms to directory
		epen.histFile.getDir("epenQ2").add("q2Raw",epen.q2Raw);
		epen.histFile.getDir("epenQ2").add("q2Acceptance",epen.q2Acceptance);
		epen.histFile.getDir("epenQ2").add("q2ep",epen.q2ep);
		epen.histFile.getDir("epenQ2").add("q2epThetaPq",epen.q2epThetaPq);
		epen.histFile.getDir("epenQ2").add("q2en",epen.q2en);
		epen.histFile.getDir("epenQ2").add("q2enThetaPq",epen.q2enThetaPq);
		epen.histFile.getDir("epenQ2").add("q2epSim",epen.q2epSim);
		epen.histFile.getDir("epenQ2").add("q2enSim",epen.q2enSim);

		// write the histograms to a hipo file
		epen.histFile.writeFile("histEpenQ2.hipo")

		// Sanity Checks ************************************************************************************************
		System.out.println("\nSanity Checks!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
		System.out.println("Accepted to Reconstructed Events	| " + 
			acceptedEvents + "/" + nEvents + " = 	" + acceptedEvents/nEvents*100 + "% reconstructed");
		System.out.println("Found Protons to Accepted Events 	| " + 
			foundProton + "/" + acceptedEvents + " = 	" + foundProton/acceptedEvents*100 + "% protons");
		System.out.println("Found Neutrons to Accepted Events 	| " + 
			foundNeutron + "/" + acceptedEvents + " = 	" + foundNeutron/acceptedEvents*100 + "% neutrons");
		System.out.println("\nProblems with the analysis:");
		System.out.println("px, py, and pz should be the same for Vector3D and Particle Classes\nThe Particle Class somehow changes the input\ncheckParticle method swims at least " + foundProton + " proton events, but this is unreproducible.")
	} // end of main method

	// Methods ##########################################################################################################
	// Table of Methods *************************************************************************************************
	// #1 | checkList = make sure we have desired Banks and Tracks (1 or 2)
	// 		#1.1 |	checkBank = check the Banks
	// 		#1.2 | 	checkTracks = check the Tracks
	// #2 | acceptanceMatching = extracts electron information and swims possible particle tracks, keeps if both swim
	// 		#2.1 | 	protonSwim = swims calculated proton track from electron info
	// 		#2.2 | 	neutronSwim = swims calculated neutron track from electron info
	// #3 |	epThetaPq = ep Block of information, passed into thetaPq method for cut // #5
	// #4 | enThetaPq = en Block of information, passed into thetaPq method for cut // #5
	// #5 | thetaPq = thetaPq cut with angle cone of < 3 degrees
	// #6 | epenSim = extract generated information from MC::Particle Bank
	// 		#6.1 | epSim = fill q2epSim histogram
	// 		#6.2 | enSim = fill q2enSim histogram

	// ******************************************************************************************************************
	// #1 Check Banks and Tracks
	boolean checkList(HipoDataEvent event){
		if(checkBank(event)) // #1.1
			if(checkTracks(event.getBank("TimeBasedTrkg::TBTracks"))) // #1.2
				return true;
		return false;
	}

	// #1.1 Check Banks
	boolean checkBank(HipoDataEvent event){
		if(event.hasBank("TimeBasedTrkg::TBTracks") && event.hasBank("ECAL::clusters")){
			return true;
		}
		return false;
	}

	// #1.2 Check Tracks
	boolean checkTracks(HipoDataBank timeBank){
		int nTracks = timeBank.rows();

		if(nTracks == 2 || nTracks ==1)
		return true;

		else
		return false;
	}

	// #2 Acceptance Matching
	boolean acceptanceMatching(HipoDataBank timeBank){ 
		int nTracks = timeBank.rows();
		int iTracks;
		int eTrack = 666;

		for(iTracks < 0; iTracks < nTracks; iTracks++){
			// extract charge information
			byte charge = timeBank.getByte("q",iTracks);

			// if we have an electron, find which track
			if(charge == -1)
			eTrack = iTracks;
			break; // if the electron is found, break out of the loop
		} // end for loop over tracks

		// if we don't have an electron, throw out event
		if(eTrack == 666)
		return false;	

		// get electron information
		float electronPx = timeBank.getFloat("p0_x",eTrack);
		float electronPy = timeBank.getFloat("p0_y",eTrack);
		float electronPz = timeBank.getFloat("p0_z",eTrack);
		Vector3D electronP = new Vector3D(electronPx, electronPy, electronPz); 
		float electronMag = electronP.mag(); // extract magnitude 
		electronP.unit(); // make unit vector
		float electronTheta = Math.acos(electronPz/electronMag);

		// calculate proton from electron info | particle:(PID, -Px, -Py, BEAM_ENERGY - Pz, Vx, Vy, Vz)
		Particle calcProtonParticle = new Particle(2212, -1*electronPx, -1*electronPy, (BEAM_ENERGY - electronPz), 0.0,0.0,0.0);
		
		// calculate neutron from electron info | Vector: (-Px, -Py, BEAM_ENERGY - Pz)
		Vector3D calcNeutronVector = new Vector3D(-1*electronPx, -1*electronPy, BEAM_ENERGY - electronPz);
		calcNeutronVector.unit(); // make unit vector

		// Q^2 Calculation
		float q2 = 4*BEAM_ENERGY*electronMag*Math.pow((Math.sin(electronTheta/2)),2);
		q2Raw.fill(q2);

		if(protonSwim(calcProtonParticle) && neutronSwim(calcNeutronVector)){ // #2.1 = protonSwim, #2.2 = neutronSwim
			q2Acceptance.fill(q2);
			// Particle Swim Sanity Checks
			// System.out.println("*****************************");
			// System.out.println("				 pid 	q 	mass 	 px 	  py 	    pz 	  	vx 	  vy 	 vz");
			// System.out.println("Calculated Proton Vector:				" + numberFormat.format(-1*electronPx) + "   " + numberFormat.format(-1*electronPy) + "   " + numberFormat.format(BEAM_ENERGY - electronPz));
			// System.out.println("Calculated Proton Particle: 	" + calcProtonParticle.toString());
			return true;
		}

		else
		return false;
	}

	// #2.1 Swim possible Proton Particle
	boolean protonSwim(Particle possibleP){
		//checkParticle Method (currently broken? Returns FTOF=0 always)
		if(recFastMC.checkParticle(possibleP) == true)
		return true;

		else
		return false;
	}

	// #2.2 Swim possible Neutron Vector
	boolean neutronSwim(Vector3D possibleN){
		boolean intersectionStatus = false;
		Line3D neutronPath = new Line3D(new Point3D(0,0,0), possibleN);
		for(int x = 0; x < 6; x++){
			intersectionStatus = ECfaces[x].hasIntersection(neutronPath);
			if(intersectionStatus == true)
			break;
		}

		if(intersectionStatus == true)
		return true;

		else
		return false;
	}

	// #3 ep Block, extract electron and proton information
	boolean epThetaPq(HipoDataBank epBank){
		int epTracks = epBank.rows();
		int eTrack;
		int pTrack;

		// check to make sure we only analyze two tracks, then distinguish between e and p
		if(epTracks == 2){
			if(epBank.getByte("q",0) == -1 && epBank.getByte("q",1) == 1){
				eTrack = 0;
				pTrack = 1;
			}
			else if(epBank.getByte("q",0) == 1 && epBank.getByte("q",1) == -1){
				eTrack = 1;
				pTrack = 0;
			}
		}

		else if(epTracks ==1)
		return false;

		// Electron Info
		float electronPx = epBank.getFloat("p0_x",eTrack);
		float electronPy = epBank.getFloat("p0_y",eTrack);
		float electronPz = epBank.getFloat("p0_z",eTrack);
		Vector3D electronP = new Vector3D(electronPx, electronPy, electronPz); 
		float electronMag = electronP.mag(); // extract magnitude 
		electronP.unit(); // make unit vector
		float electronTheta = Math.acos(electronPz/electronMag);

		// Q^2 Calculation
		float q2 = 4*BEAM_ENERGY*electronMag*Math.pow((Math.sin(electronTheta/2)),2);
		q2ep.fill(q2);

		// Proton Info
		float protonPx = epBank.getFloat("p0_x",pTrack);
		float protonPy = epBank.getFloat("p0_y",pTrack);
		float protonPz = epBank.getFloat("p0_z",pTrack);
		Vector3D protonP = new Vector3D(protonPx, protonPy, protonPz); 
		protonP.unit(); // make unit vector

		// calculate proton vector from electron info | Vector: (-Px, -Py, BEAM_ENERGY - Pz)
		Vector3D calcProtonVector = new Vector3D(-1*electronPx, -1*electronPy, BEAM_ENERGY - electronPz);
		calcProtonVector.unit(); // make unit vector

		// pass info into thetaPq cut method #5
		if(thetaPq(protonP, calcProtonVector)){
			q2epThetaPq.fill(q2);
			return true;
		} 

		else
		return false;
	} // end of ep Block boolean

	// #4 en Block, extract electron and neutron information
	boolean enThetaPq(HipoDataBank enTimeBank, HipoDataBank enEcBank){
		int enTracks = enTimeBank.rows();
		int eTrack;
		int numThetaPq;

		if(enTracks == 1)
			if(enTimeBank.getByte("q",0) == -1)
			eTrack = 0;

		else 
		return false;

		// Electron Info
		float electronPx = enTimeBank.getFloat("p0_x",eTrack);
		float electronPy = enTimeBank.getFloat("p0_y",eTrack);
		float electronPz = enTimeBank.getFloat("p0_z",eTrack);
		Vector3D electronP = new Vector3D(electronPx, electronPy, electronPz); 
		float electronMag = electronP.mag(); // extract magnitude 
		electronP.unit(); // make unit vector
		float electronTheta = Math.acos(electronPz/electronMag);

		// Q^2 Calculation
		float q2 = 4*BEAM_ENERGY*electronMag*Math.pow((Math.sin(electronTheta/2)),2);
		q2en.fill(q2);

		// calculate neutron vector from electron info | Vector: (-Px, -Py, BEAM_ENERGY - Pz)
		Vector3D calcNeutronVector = new Vector3D(-1*electronPx, -1*electronPy, BEAM_ENERGY - electronPz);
		calcNeutronVector.unit(); // make unit vector

		// Neutron Info
		int ecRows = enEcBank.rows();
		for(int n = 0; n < ecRows; n++){
			float hitX = enEcBank.getFloat("x",n);
			float hitY = enEcBank.getFloat("y",n);
			float hitZ = enEcBank.getFloat("z",n);
			Vector3D neutronP = new Vector3D(hitX, hitY, hitZ);
			neutronP.unit(); // make possible neutron vector a unit vector

			if(thetaPq(neutronP, calcNeutronVector)){
				q2enThetaPq.fill(q2);
				return true;
				break;
			}

			else
			return false;

		} // end for loop searching for hits in ECAL clusters
	} // end of en Block boolean

	// #5 thetaPq cut
	boolean thetaPq(Vector3D foundVector, Vector3D calcVector){
		float thetaPq = Math.acos(calcVector.dot(foundVector))*(180/Math.PI); 
		if(thetaPq <= 3){ // theta cut
			return true;
		}

		else 
		return false;
	}

	// #6 Q^2 from generated bank
	void epenSim(HipoDataBank mcBank){
		int epTracks = mcBank.rows();
		int eTrack;
		int pTrack;
		int nTrack;

		// check to make sure we only analyze two tracks, then distinguish between e and p
		if(epTracks == 2){
			if(mcBank.getInt("pid",0) == 11 && mcBank.getInt("pid",1) == 2212){
				eTrack = 0;
				pTrack = 1;
			}
			else if(mcBank.getInt("pid",0) == 11 && mcBank.getInt("pid",1) == 2112){
				eTrack = 0;
				nTrack = 1;
			}
		}

		// Electron Info
		float electronPx = mcBank.getFloat("px",eTrack);
		float electronPy = mcBank.getFloat("py",eTrack);
		float electronPz = mcBank.getFloat("pz",eTrack);
		Vector3D electronP = new Vector3D(electronPx, electronPy, electronPz); 
		float electronMag = electronP.mag(); // extract magnitude 
		electronP.unit(); // make unit vector
		float electronTheta = Math.acos(electronPz/electronMag);

		// Q^2 Calculation
		float q2 = 4*BEAM_ENERGY*electronMag*Math.pow((Math.sin(electronTheta/2)),2);
		
		if(pTrack == 1)
		epSim(q2);

		else if(nTrack == 1)
		enSim(q2);
	} // end of epenSim Method

	// #6.1 fill e-p Generated Q^2 hist
	void epSim(float q2){
		q2epSim.fill(q2);
	}

	// #6.2 fill e-n Generated Q^2 hist
	void enSim(float q2){
		q2enSim.fill(q2);
	}
} // end of class QuasiElastic